// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{fs, path::Path};

use glium::{Surface, VertexBuffer, Display, Program, Frame, DrawParameters, index::{NoIndices, PrimitiveType}, uniforms::EmptyUniforms, implement_vertex};
use glutin::surface::WindowSurface;
use glam::Vec2;

use super::{Camera, star::Star};

use crate::game::filesystem::{Filesystem, FsBase};
use crate::utils::glium_resize;

#[derive(Clone, Copy)]
struct Vertex { v_pos: [f32; 2], v_alpha: f32 }
implement_vertex!(Vertex, v_pos, v_alpha);

pub(super) struct Stars {
	stars: Vec<Star>,
	density: f32,
	c_size: Vec2,

	vbo: VertexBuffer<Vertex>,
	program: Program,
}

impl Stars {
	pub fn new(display: &Display<WindowSurface>, fs: &Filesystem) -> Stars {
		let vsh = fs::read_to_string(fs.get(FsBase::Static, Path::new("shaders/stars.vsh"))).unwrap();
		let fsh = fs::read_to_string(fs.get(FsBase::Static, Path::new("shaders/stars.fsh"))).unwrap();

		Stars {
			stars: Vec::new(),
			density: 0.0,
			c_size: Vec2::ZERO,
			vbo: VertexBuffer::empty_persistent(display, 0).unwrap(),
			program: Program::from_source(display, &vsh, &fsh, None).unwrap(),
		}
	}

	pub fn reset(&mut self, camera: &Camera, density: f32) {
		self.stars.clear();
		self.density = density;
		self.c_size = camera.get_size();
	}

	pub fn update(&mut self, dt: f32, camera: &Camera, to_move: Vec2) {
		let new_c_size = camera.get_size();

		let moved_vel = to_move / dt;

		for star in &mut self.stars {
			if new_c_size != self.c_size {
				star.resize(self.c_size, new_c_size);
			}
			star.move_by(to_move);
			star.update(dt);
			star.out_of_bounds(camera, moved_vel);
		}

		self.c_size = new_c_size;

		/*
		 * Adds or removes stars if needed to keep the star density constant as
		 * the window area changes.
		 *
		 * Note that this must be called after moving the stars to prevent
		 * `Star::resize` being called on an already resized star, which results
		 * in the new stars being spawned in a non-uniform distribution.
		 */
		let pixels = camera.get_window_size().element_product();
		let expected_len = (self.density / 1e6 * pixels as f32).round() as usize;

		if self.stars.len() < expected_len {
			let count = expected_len - self.stars.len();
			self.stars.reserve(count);
			for _ in 0..count {
				self.stars.push(Star::new(camera));
			}
		} else {
			self.stars.truncate(expected_len);
		}
	}

	pub fn render(&mut self, display: &Display<WindowSurface>, frame: &mut Frame, params: &DrawParameters) {
		glium_resize::vbo_persistent(&mut self.vbo, display, self.stars.len());

		let zoom = self.c_size.min_element();
		{
			let mut buf = self.vbo.map_write();
			for (i, star) in self.stars.iter().enumerate() {
				let fade_out = ((1.0 - i as f32 / self.stars.len() as f32) * 3.0).min(1.0);
				buf.set(i, Vertex { v_pos: (star.get_rendered_pos(zoom) / self.c_size * 2.0).into(), v_alpha: star.get_alpha(zoom) * fade_out });
			}
		}

		frame.draw(self.vbo.slice(..self.stars.len()).unwrap(), NoIndices(PrimitiveType::Points), &self.program, &EmptyUniforms, params).unwrap();
	}
}
